from tensorflow.keras import Model
import tensorflow as tf

def loss(margin=1):
    """Provides 'constrastive_loss' an enclosing scope with variable 'margin'.

    Arguments:
        margin: Integer, defines the baseline for distance for which pairs
                should be classified as dissimilar. - (default is 1).

    Returns:
        'constrastive_loss' function with data ('margin') attached.
    """

    # Contrastive loss = mean( (1-true_value) * square(prediction) +
    #                         true_value * square( max(margin-prediction, 0) ))
    def contrastive_loss(y_true, y_pred):
        """Calculates the constrastive loss.

        Arguments:
            y_true: List of labels, each label is of type float32.
            y_pred: List of predictions of same length as of y_true,
                    each label is of type float32.

        Returns:
            A tensor containing constrastive loss as floating point value.
        """

        square_pred = tf.math.square(y_pred)
        margin_square = tf.math.square(tf.math.maximum(margin - (y_pred), 0))
        return tf.math.reduce_mean(
            (1 - y_true) * square_pred + (y_true) * margin_square
        )

    return contrastive_loss

def euclidean_distance(vects):
    """Find the Euclidean distance between two vectors.

    Arguments:
        vects: List containing two tensors of same length.

    Returns:
        Tensor containing euclidean distance
        (as floating point value) between vectors.
    """

    x, y = vects
    sum_square = tf.math.reduce_sum(tf.math.square(x - y), axis=1, keepdims=True)
    return tf.math.sqrt(tf.math.maximum(sum_square, tf.keras.backend.epsilon()))
def Siamese_con(kwargs):
    input,x=get_model_rd_con(**kwargs)
    embedding_network = Model(input, x)


    input_1 = tf.keras.layers.Input((64, 2))
    input_2 = tf.keras.layers.Input((64, 2))

    # As mentioned above, Siamese Network share weights between
    # tower networks (sister networks). To allow this, we will use
    # same embedding network for both tower networks.
    tower_1 = embedding_network(input_1)
    tower_2 = embedding_network(input_2)

    normal_layer = tf.keras.layers.Lambda(euclidean_distance)([tower_1, tower_2])
    output_layer = tf.keras.layers.Dense(1, activation="sigmoid")(normal_layer)
    return Model(inputs=[input_1, input_2], outputs=output_layer)

def get_model_rd_con(hidden1=[64],hidden2=[64],dropout=0.0,batchnorm=False):
    input = tf.keras.layers.Input(shape=(64, 2))
    x1 = tf.keras.layers.Lambda(lambda x: x[..., 0])(input)
    x1=tf.keras.layers.Reshape((1, 64))(x1)
    for h in hidden1:
        x1 = tf.keras.layers.Dense(h, activation="relu")(x1)
        if dropout>0.0:
            x1=tf.keras.layers.Dropout(dropout)(x1)
        if batchnorm==True:
            x1=tf.keras.layers.BatchNormalization()(x1)
    x2 = tf.keras.layers.Lambda(lambda x: x[..., 1])(input)
    x2=tf.keras.layers.Reshape((1, 64))(x2)
    for h in hidden1:
        x2 = tf.keras.layers.Dense(h, activation="relu")(x2)
        if batchnorm==True:
            x2=tf.keras.layers.BatchNormalization()(x2)
    x= tf.keras.layers.Concatenate()([x1,x2])
    for h in hidden2:
        x = tf.keras.layers.Dense(h, activation="relu")(x)
        if batchnorm==True:
            x=tf.keras.layers.BatchNormalization()(x)
    return input,x
